(* Example usage: ./conjecture *)

open Init
open Hh_parse
open Toolbox
open Read
open Pattern
open Matching_multiple
open Multiple_experiments
open Matching_subst
open Matching_translate

let eval_file_glob = ref ""

(* Parse command-line arguments, exiting with help if not valid.
   TODO: Read a conjecture itself, not its name. *)
let parse_commandline () =
  no_congruence_flag := true;
  (* Saving results into two references and a array. *)
  let anon_counter = ref 0 in
  let anon_tab = Array.make 2 "" in
  let parse_anon anon = anon_tab.(!anon_counter) <- anon; incr anon_counter in
  let speclist =
    [("-eval", Arg.Set_string eval_file_glob, 
      "\nImport the list of theorems' name from which conjectures will be produced." ^ 
      "\nIf no evaluation file is given, all theorems in the library will be used.");
     ("-subterm", Arg.Set subterm_flag,  
      "\nAllow frequent higher-order ground subterms to be matched." );
     ("-fof",  Arg.Set fof_flag,  "\nOutput conjectures in fof format instead of our internal 
     format tt. Only use when input formulas are first-order." );
   ]
  in
  let usage_msg = "Conjecture-maker. Usage: " ^ Sys.argv.(0) ^ " <dir_in> <dir_out>" in
  Arg.parse speclist parse_anon usage_msg;
  (* Processing the parsed objects. *)
  let dir_in = anon_tab.(0) in
  let dir_out = anon_tab.(1) in
  (dir_in, dir_out)


(* Main function. *)
let _ =
  let (dir_in, dir_out) = parse_commandline () in
  if !eval_file_glob = "" then () else init_eval_glob !eval_file_glob;
  make_conjectures_dir dir_in dir_out
;;
